#ifndef WMMSENS_CHIPSET_H
#define WMMSENS_CHIPSET_H

//
// chipset.h - new implementation of sensor manager classes
//   by Ed Porras <ep@bangchemistry.com>
//   with portions of code from
//
//     wmgtemp by Roger Dunce <kronos@nanunanu.org> 
//
//
// $Log: chipset.h,v $
// Revision 1.1  2003/03/25 02:14:20  ep
// Initial revision
//
// 

#include <iosfwd>
#include <string>
#include <vector>

#include <sensors/sensors.h>
#include "cs_feature.h"

namespace wmmsens
{
  //
  // stores all features for a specific chipset
  class chipset
  {
  public:
    enum csf_type { 
      POLL_TEMP      = 0x01,
      POLL_FAN       = 0x02,
      POLL_VOLTAGE   = 0x04, 
    };

    ~chipset();

  private:
    // only chipset_mgr can use this class..
    friend class chipset_mgr;

    // the chipset struct.. this is a copy of the one returned by the
    // sensor lib..
    sensors_chip_name chipname;

    // list of non-mapped features for this chipset
    std::vector<cs_feature *> feature_list;

    ui16 poll_bits;

    // prohibit
    chipset( const chipset& );
    chipset& operator=( const chipset& );

    // constructor
    chipset(const sensors_chip_name& name, ui16 bits);

    // getters
    std::string name() const { return chipname.prefix; }
    ui32 numFeatures() const { return feature_list.size(); }

    // get the chip's features.. history steps refers to the step
    // count for history items (currently only temp uses this)
    bool getFeatures(ui8 history_steps);

    // update all readings
    void update(bool all = false); 

    bool storeMappedFeature(const sensors_feature_data& sfd);

    // fetches the feature at the given index
    const cs_feature& operator[](ui32 idx) const {
      return *feature_list[ idx %= feature_list.size() ];
    }

    friend std::ostream& operator<<(std::ostream &, const chipset&);
  };

} // namespace wmmsens

#endif
