//
// multiple sensor display dock app by Ed Porras <ep@bangchemistry.com>.
// based on:
//
//     wmgtemp by Roger Dunce <kronos@nanunanu.org> 
//
// with scrolling text routines from:
//
//     wmix by Daniel Richard G. <skunk@mit.edu> / 
//             timecop <timecop@japan.co.jp>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// $Log: main.c,v $
// Revision 1.25  2003/03/25 02:14:59  ep
// adapted to new classes & for compilation under gcc3
//
// Revision 1.24  2001/11/16 04:19:23  ep
// implemented new legend mechanism where a bullet precedes
// the displayed info.. changed scrolltext to use bools..
//
// Revision 1.23  2001/11/13 07:28:27  ep
// changed legend indicator to horiz line.. added parens to
// xpm
//
// Revision 1.22  2001/11/13 02:04:25  ep
// added suport for deg kelvin
//
// Revision 1.21  2001/11/12 23:48:11  ep
// preliminary implementation of graph legend
//
// Revision 1.20  2001/11/12 08:12:00  ep
// added --all option.. bumped to 0.29.4
//
// Revision 1.19  2001/11/09 07:40:27  ep
// changed command-line args.. fixed dumb issue with progname
//
// Revision 1.18  2001/11/08 09:45:13  ep
// adapted to new pixmap.. units are now set here..
//
// Revision 1.17  2001/11/07 08:43:04  ep
// changed --f flag to --ft.. disabled fan polling as a default
// so now flag has to be passed to enable it.. cpu voltages are
// turned on with flag but scroll text must be fixed for readability
//
// Revision 1.16  2001/11/06 06:53:12  ep
// adapted to use new manager classes.. handles args for
// fahrenheit & verbose..
// preliminary for 0.30 (major planned release)
//
// Revision 1.15  2001/11/05 08:05:05  ep
// incremental save while finishing port to new chipset management
// classes
//
// Revision 1.14  2001/11/04 22:55:10  ep
// added code for displaying temperature in C or F.. need
// mechanism for allowing user to set the mode
//
// Revision 1.13  2001/11/01 23:20:07  ep
// bumped version
//
// Revision 1.12  2001/10/31 08:00:51  ep
// fixed scrolling jump problem (got rid of alarm)..
//
// Revision 1.11  2001/10/30 23:27:19  ep
// fixed graph issues where points were being drawn out of range
// and therefore not cleared
//
// Revision 1.10  2001/10/20 06:27:29  ep
// enabled range lines and renamed a few thigns for consistency
//
// Revision 1.9  2001/10/19 01:51:34  ep
// minor mods dealing with constant names & comments
//
// Revision 1.8  2001/10/18 20:41:33  ep
// got rid of harcoded recognized types and changed to
// use chipset detecting routines
//
// Revision 1.7  2001/10/18 18:49:43  ep
// minor irrelevant fixes
//
// Revision 1.6  2001/10/18 18:03:18  ep
// added version info and some arg parse code
//
// Revision 1.5  2001/10/18 17:40:20  ep
// modified enums according to new xpm.. made some more relative ones..
// implemented red-text in case limit is reached by feature
//
// Revision 1.4  2001/10/18 03:05:22  ep
// fixed file close in initsensors (in case file is not found)..
// minor changes to output
//
// Revision 1.3  2001/10/17 17:47:01  ep
// renamed some enums for consistency.. adapted to use feature-detecting
// code in sensor_mgr
//
// Revision 1.2  2001/10/17 15:00:18  ep
// added scrolling text routines from wmix
// removed obsolete code..
//
// Revision 1.1  2001/10/16 15:43:41  ep
// Initial revision
//
//

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cstring>
#include <unistd.h>
#include <sensors/sensors.h>
#include <sensors/chips.h>
#include <sensors/error.h>
#include "wmgeneral/wmgeneral.h"
#include "wmmsens-interface.xpm"
#include "wmmsens-interface-mask.xbm"
#include "chipset_mgr.h"

using namespace std;

//
// the "globals" we need
namespace wmmsens
{
    const string VER = "0.29.6";
    const string AUTHOR = "Ed Porras <ep@bangchemistry.com>";
    const string OTHER = 
    "  based on wmgtemp by Roger Dunce <kronos@nanunanu.org>\n" \
    "  with scrolling text routines based on code from\n" \
    "  wmix by Daniel Richard G. <skunk@mit.edu> / " \
    "timecop <timecop@japan.co.jp>\n" \
    "  http://www.elguapos.net/wmmsens/\n";

    // more constants
    enum { 
	DOCKAPP_DIM             = 64,
	BORDER_WIDTH            = 3,
	HISTORY_STEPS           = DOCKAPP_DIM - (2 * BORDER_WIDTH),  

	DISPLAY_MIN             = 20,
	DISPLAY_MAX             = 40,

	XPM_CHAR_WIDTH          = 6, 
	XPM_CHAR_2XWIDTH        = XPM_CHAR_WIDTH * 2, 
	XPM_CHAR_HEIGHT         = 7,
	XPM_LINE_HEIGHT         = XPM_CHAR_HEIGHT + 2,
	XPM_LINE_2XHEIGHT       = XPM_LINE_HEIGHT * 2,

	XPM_TEXT_TOP            = DOCKAPP_DIM - BORDER_WIDTH - 
	                          XPM_LINE_HEIGHT,
	XPM_TEXT_X              = BORDER_WIDTH, 
	XPM_TEXT_WIDTH          = HISTORY_STEPS,

	XPM_GRAPH_TOP           = BORDER_WIDTH,
	XPM_GRAPH_BOTTOM        = XPM_TEXT_TOP - 3,
	XPM_GRAPH_HEIGHT        = XPM_GRAPH_BOTTOM - XPM_GRAPH_TOP - 1,
	XPM_GRAPH_X             = XPM_TEXT_X,

	XPM_DATA_VLINE_Y        = XPM_GRAPH_TOP,
	XPM_DATA_CLEAR_VLINE_X  = 65,
	XPM_DATA_VLINE_X_OFFSET = XPM_DATA_CLEAR_VLINE_X + 1,

	XPM_LEGEND_Y            = 50, 
	XPM_LEGEND_WIDTH        = 4,
	XPM_LEGEND_HEIGHT       = 6,
	XPM_LEGEND_X            = XPM_DATA_CLEAR_VLINE_X + XPM_LEGEND_WIDTH,

	XPM_NONALPHA_LINE_Y     = 64,
	XPM_RED_NONALPHA_LINE_Y = XPM_NONALPHA_LINE_Y + XPM_LINE_HEIGHT,
	XPM_ALPHA_LINE_Y        = XPM_RED_NONALPHA_LINE_Y + XPM_LINE_HEIGHT,
	XPM_RED_ALPHA_LINE_Y    = XPM_ALPHA_LINE_Y + XPM_LINE_HEIGHT,
	XPM_BLANK_LINE_Y        = XPM_RED_ALPHA_LINE_Y + XPM_LINE_HEIGHT,
	XPM_WRITE_LINE_Y        = XPM_BLANK_LINE_Y + XPM_LINE_HEIGHT,
	XPM_LINE_X              = 0,
	XPM_DEGREE_X            = 168, 
	XPM_DEG_C_X             = XPM_DEGREE_X + XPM_CHAR_WIDTH,

	XPM_VOLT_X              = 192,
	XPM_RPM_X               = XPM_VOLT_X + XPM_CHAR_WIDTH,

	XPM_RANGE_LINE_X        = 80,
	XPM_U_RANGE_LINE_Y      = 0,
	XPM_L_RANGE_LINE_Y      = XPM_U_RANGE_LINE_Y + 1,
    };

    // temperature display mode
    enum t_mode { C, F, K };
    t_mode display_mode = C; // default is celcius

    ui16 poll = chipset::POLL_TEMP;

    // these should be obvious..
    const featval F_FREEZE = 32.0;
    const featval F_DEG_RAT = 9.0 / 5.0;
    const featval K_OFFSET = 273.15;

    // vertical range of the history 
    double range_upper = DISPLAY_MAX;
    double range_lower = DISPLAY_MIN;
    double range_step  = 5.0;

    // the instance of the manager keeping track of things
    chipset_mgr cs_mgr(HISTORY_STEPS);

    // delay for drawing 
    int delay = 20; // 1/10ths of a second

    // duh..
    std::string progname;

  //
  // protos
  bool initSensors();
  void processXevents();
  int blitString(const string& text, bool warn = false);
  bool scrollText(int x, int y, int width, int ctlength, bool reset);


  //
  // no usage really yet
  void usage()
  {
    cout << "Usage: " << progname << " [option list]" << endl
	 << "  where, [options list] can consist of: " << endl
	 << "     --help:    this screen" << endl
	 << "     --version: display program version info" << endl
	 << "     --verbose: be verbose" << endl
	 << "     --degf:    use the fahrenheit scale" << endl
	 << "     --degk:    use the kelvin scale" << endl
	 << "     --volt:    monitor cpu voltage readings" << endl
	 << "     --fan:     monitor fan readings" << endl
	 << "     --all:     monitor temperature, voltage, & fan readings" 
	 << endl;
  }


  //
  // initializes the sensor system
  bool initSensors() 
  {
    // initialize sensors using sensors config file (look for it)
    FILE* config_file = fopen("/etc/sensors.conf", "r");
    int res = sensors_init(config_file);

    if (res != 0) {
      config_file = fopen("/usr/etc/sensors.conf", "r");
      res = sensors_init(config_file);

      if (res != 0) {
	config_file = fopen("/usr/local/etc/sensors.conf", "r");
	res = sensors_init(config_file);

	if (res != 0) {
	  cerr << sensors_strerror(res) << endl;
	  return false;
	}
      }
    }

    // don't close it unless it was actually opened..
    if (config_file != NULL)
      fclose(config_file);

    // try to detect chipsets with standard feature names.. here's
    // where things like lm80, etc. fail because they use names like
    // temp_hot_max, etc.. it's a pain to guess all the different
    // ones..
    if (!cs_mgr.detect(poll))
    {
      cerr << progname << " failed to detect a standard chipset" << endl;
      return false;
    }

    return true;
  }

  //
  // determine the range of temperatures we need to be drawing.. 
  bool recomputeRange()
  {
    bool modified = false;
    bool pass_high = true, pass_low = true;

    // check everyone of the entries polled
    for (register ui32 i = 0; i < cs_mgr.numFeatures(); i++)
    {
      const cs_feature& f = cs_mgr[i];

      // only use this entry if we're keeping a history of it
      if (!f.history())
	continue;

      // if the stored high and low is greater / lower than the range,
      // adjust the range
      if (f.max() > range_upper)
      {
	range_upper += range_step;
	modified = true;
      }
      if (f.min() < range_lower) 
      { 
	range_lower -= range_step;  
	modified = true;  
      }  

      // determine whether it passed the high / low point
      pass_high &= (f.max() < (range_upper - range_step) &&
		    (range_upper - range_step) >= DISPLAY_MAX);
      pass_low &= (f.min() > (range_lower + range_step) &&
		   (range_lower + range_step) <= DISPLAY_MIN );
    }

    if (pass_high) 
    {
      range_upper -= range_step;
      modified = true;
    }
    if (pass_low)
    {
      range_lower += range_step;  
      modified = true;
    }  
    return modified;
  }


  //
  // draw one of the visible min / max lines..
  inline void drawRangeLine(double temp, double step, short type) 
  {
    ui32 height = static_cast<ui32>(step * (temp - range_lower));

    copyXPMArea(XPM_RANGE_LINE_X, type, HISTORY_STEPS, 1, 
		XPM_GRAPH_X, XPM_GRAPH_HEIGHT - height + XPM_GRAPH_TOP);
  }

  //
  // updates the graphical display
  void updateGraph() 
  {
    // rescale the display if needed.
    while (recomputeRange());

    // compute the amount each unit occupies on the graph
    double range = range_upper - range_lower;
    double h_ratio = XPM_GRAPH_HEIGHT / range;

    // redraw the graph
    for (register ui32 i = 0; i < HISTORY_STEPS; i++) 
    {
      // clear a line
      copyXPMArea( XPM_DATA_CLEAR_VLINE_X, XPM_DATA_VLINE_Y, 
		   1, XPM_GRAPH_HEIGHT + 1, i + XPM_GRAPH_X, XPM_GRAPH_TOP );

      // draw the temperatures on the graph.
      for (register ui32 j = 0, k = 0; j < cs_mgr.numFeatures(); j++)
      {
	const cs_feature& f = cs_mgr[j];

	if (f.history())
	{
	  // if there's no recorded value (it's approx 0), skip it
	  double temp = f.value(i);
	  if (temp < 0.1)
	    continue;

	  // otherwise, draw a dot
	  ui32 height = static_cast<ui32>(h_ratio * (temp - range_lower));
	  copyXPMArea( XPM_DATA_VLINE_X_OFFSET + k++, XPM_DATA_VLINE_Y, 
		       1, 1, 
		       i + XPM_GRAPH_X, (XPM_GRAPH_HEIGHT - height) + XPM_GRAPH_TOP);
	}
      }
    }

    // draw range lines if needed
    if (range_upper > DISPLAY_MAX)
      drawRangeLine(DISPLAY_MAX, h_ratio, XPM_U_RANGE_LINE_Y);
    if (range_lower < DISPLAY_MIN)
      drawRangeLine(DISPLAY_MIN, h_ratio, XPM_L_RANGE_LINE_Y);
  }


  //
  // convenience
  inline bool scrollText(int ctlength, bool reset)
  {
    return scrollText(XPM_TEXT_X, XPM_TEXT_TOP, XPM_TEXT_WIDTH, 
		      ctlength, reset);
  }

  //
  // window events
  void processXevents() 
  {
    XEvent Event;
    ui32 cur_feat = 0, hist_idx = 0;
    bool reset = false;
    int mark = -1;

    // start scrolling the prog name
    int text_pix_w = blitString(progname + " " + VER);
    scrollText(text_pix_w, true);

    while (true) 
    {
      while (XPending(display)) 
      {
	XNextEvent(display, &Event);

	switch (Event.type) 
	{
	case Expose:
	  RedrawWindow();
	  break;

	case DestroyNotify:
	  XCloseDisplay(display);
	  return;
	  break;

	default:
//	       cout << " -*- unhandled event: " << Event.type << endl;
	  break;
	}
      }

      // sleep for 1/10th of a second
      usleep(100000);

      // update the counter 
      mark = (mark + 1) % delay;

      // if it's been two seconds, read the sensor data
      if (mark == 0)
      {
	// update sensor data and the display
	cs_mgr.update();
	updateGraph();
      }

      // scroll text if we have to
      if (!scrollText(text_pix_w, reset))
      {
	ostringstream line;

	const cs_feature& cur = cs_mgr[ cur_feat ];

	if (cur.history())
	  line << "\b" << hist_idx++;

	// format the line with label..
	line << cur.label() << ": ";

	// and value.. convert it if necessary
	featval val = cur.value();

	// output with the corresponding unit
	switch (cur.type())
	{
	case cs_feature::TEMPERATURE:
	  // convert it if requested
	  switch (display_mode)
	  {
	  case F:
	    // convert to fahrenheit..
	    val = (val * F_DEG_RAT) + F_FREEZE;
	    break;

	  case K:
	    // convert to kelvin
	    val += K_OFFSET;
	    break;

	  default:
	    // celcius is the default (lm_sensors returns
	    // celcius values)
	    break;
	  }

	  // mapped to deg symbol
	  line << setprecision(3) << val << "\t";
	  break;

	case cs_feature::FAN:
	  // mapped to rpm symbol
	  line << val;// << "\f"; // disabled for now
	  break;
	       
	case cs_feature::VOLTAGE:
	  // mapped to small voltage symbol
	  line << setprecision(3) << val;// << "\v"; // disabled for now
	  break;
	}

	// build the string..
	text_pix_w = blitString( line.str(), cur.alarm() );

	// force scroll to start on next pass
	reset = true;

	// the sensor data we're currently scrolling is marked by
	// cur_feat.. 
	cur_feat++;
	if (cur_feat == cs_mgr.numFeatures())
	{
	  cur_feat %= cs_mgr.numFeatures();
	  hist_idx = 0;
	}
      }
      else reset = false;
    }
  }

  //
  // based on code from wmix.. minor mods to make it more dynamic and
  // optimize it a bit.. write the string using the xpm characters into
  // a predefined blank area of the xpm.. we'll then use the area to
  // write the text on the app.. returns the width of the resulting text
  // in pixels
  int blitString(const string& text, bool warn)
  {
    register int c, k = 0;

    // clear the area..
    copyXPMArea(0, XPM_BLANK_LINE_Y, 256, XPM_LINE_HEIGHT, 0, 
		XPM_WRITE_LINE_Y);

    int src_y = XPM_NONALPHA_LINE_Y + (warn ? XPM_LINE_HEIGHT : 0);

    // go through char by char
    for (register ui32 i = 0; i < text.length() || i > 31; i++) 
    {
      c = text[i];

      switch (c)
      {
      case '\b':
	if ((i + 1) < text.length()) // paranoia
	{
	  // add a legend indicator for graphed features.. we use
	  // \b to mark the spot where it should be
	  // inserted.. the digit following it indicates the line
	  // number drawn so we can use the same color.. this
	  // assumes what follows the \b is a digit value!
	  int line_no = text[++i] - '0';
	  copyXPMArea(XPM_LEGEND_X + (line_no * XPM_LEGEND_WIDTH), 
		      XPM_LEGEND_Y,
		      XPM_LEGEND_WIDTH, XPM_LEGEND_HEIGHT, 
		      k, XPM_WRITE_LINE_Y);
	  k += XPM_CHAR_WIDTH;
	}
	break;

      case '\t':
	// mapped to degree symbol
	copyXPMArea(XPM_DEGREE_X, src_y,
		    XPM_CHAR_WIDTH, XPM_CHAR_HEIGHT, 
		    k, XPM_WRITE_LINE_Y);
	k += XPM_CHAR_WIDTH;

	// add the smaller c or f for centigrade or fahrenheit
	copyXPMArea(XPM_DEG_C_X + (XPM_CHAR_WIDTH * display_mode), 
		    src_y,
		    XPM_CHAR_WIDTH, XPM_CHAR_HEIGHT, 
		    k, XPM_WRITE_LINE_Y);
	k += XPM_CHAR_WIDTH;
	break;

      case '\v':
	// mapped to small V
	copyXPMArea(XPM_VOLT_X, src_y,
		    XPM_CHAR_WIDTH, XPM_CHAR_HEIGHT, 
		    k, XPM_WRITE_LINE_Y);
	k += XPM_CHAR_WIDTH;
	break;

      case '\f':
	// mapped to rpm
	copyXPMArea(XPM_RPM_X, src_y,
		    XPM_CHAR_2XWIDTH + 1, XPM_CHAR_HEIGHT, 
		    k, XPM_WRITE_LINE_Y);
	k += XPM_CHAR_2XWIDTH;
	break;

      default:
	if (c >= ' ' && c <= ';')
	{
	  c = c - ' ';
	  copyXPMArea(c * XPM_CHAR_WIDTH, src_y, 
		      XPM_CHAR_WIDTH, XPM_CHAR_HEIGHT, 
		      k, XPM_WRITE_LINE_Y);
	  k += XPM_CHAR_WIDTH;
	}
	else
	{
	  // either alpha or unhandled
	  c = toupper(text[i]);
	  if (c >= 'A' && c <= 'Z') {
	    c = c - 'A';
	    copyXPMArea(c * XPM_CHAR_WIDTH, src_y + XPM_LINE_2XHEIGHT,
			XPM_CHAR_WIDTH, XPM_CHAR_HEIGHT, 
			k, XPM_WRITE_LINE_Y);
	    k += XPM_CHAR_WIDTH;
	  }
	}
      }
    }

    // return the width in pixels
    return k;
  }


  //
  // based on code (well, this one is almost exactly) from wmix.. reset
  // indicates whether to start from scratch or not
  bool scrollText(int x, int y, int width, int ctlength, bool reset)
  {
    static int pos = 0;
    static bool first = false;
    static bool stop = false;

    if (reset) {
      // restart the scrolling..
      pos = 0;
      first = stop = false;
      copyXPMArea(0, XPM_BLANK_LINE_Y, width, XPM_CHAR_HEIGHT, x, y);
    }

    // if we've reached the end, do nothing..
    if (stop) {
      return false;
    }

    // otherwise, set starting point to the right side..
    if ((!first) && pos == 0) {
      pos = width;
      first = true;
    }

    // and if we reach the left side, stop moving that way..
    if (pos < -(ctlength)) {
      first = true;
      pos = width;
      stop = true;
      return false;
    }

    // scroll step
    pos -= 2;

    // draw 
    if (pos > 0) {
      // clear first.. 
      copyXPMArea(0, XPM_BLANK_LINE_Y, pos, XPM_CHAR_HEIGHT, x, y);
      copyXPMArea(0, XPM_WRITE_LINE_Y, width - pos, XPM_CHAR_HEIGHT, 
		  x + pos, y);
    } else {
      // don't need to clear, already in text
      copyXPMArea(abs(pos), XPM_WRITE_LINE_Y, width, XPM_CHAR_HEIGHT, x, y);
    }
    RedrawWindow();
    return true;
  }

} // namespace wmmsens

//
// it all starts here
int main(int argc, char **argv) 
{
  // save the program name for later use
  wmmsens::progname = argv[0];

  // if it was run with a path (e.g.: ./name), get rid of the path
  // name
  int slash = wmmsens::progname.rfind("/");
  if (slash != -1)
    wmmsens::progname = wmmsens::progname.substr(slash + 1, 
						 wmmsens::progname.length());

  // initial version of parse args
  for (register int i = 1; i < argc; i++)
  {
    std::string arg = argv[i];
    if (arg == "--version")
    {
      cout << wmmsens::progname << " " << wmmsens::VER 
	   << " by " << wmmsens::AUTHOR << endl
	   << wmmsens::OTHER << endl;
      return 0;
    }
    else if (arg == "--help")
    {
      wmmsens::usage();
      return 0;
    }
    else if (arg == "--verbose")
    {
      // enable messages to cout
      wmmsens::chipset_mgr::verbose = true;
    }
    else if (arg == "--degf")
    {
      // display temperature in fahrenheit degrees
      wmmsens::display_mode = wmmsens::F;
    }
    else if (arg == "--degk")
    {
      // display temperature in kelvin degrees.. npt sure if anyone
      // will ever use this, but..
      wmmsens::display_mode = wmmsens::K;
    }
    else if (arg == "--volt")
    {
      wmmsens::poll |= wmmsens::chipset::POLL_VOLTAGE;
    }
    else if (arg == "--fan")
    {
      wmmsens::poll |= wmmsens::chipset::POLL_FAN;
    }
    else if (arg == "--all")
    {
      wmmsens::poll |= wmmsens::chipset::POLL_FAN | 
	wmmsens::chipset::POLL_VOLTAGE;
    }
    else
    {
      cerr << "Unrecognized argument " << arg << endl << endl;
      wmmsens::usage();
      return -1;
    }
  }

  // try to setup lm sensors
  if (!wmmsens::initSensors()) 
  {
    cerr << wmmsens::progname << ": Error initializing lm_sensors" << endl;
    return -1;
  }

  // create and draw the display window
  openXwindow(argc, argv, wmmsens_interface_xpm, 
	      wmmsens_interface_mask_bits,
	      wmmsens_interface_mask_width, wmmsens_interface_mask_height);
  RedrawWindow();
    
  // now loop infinitely handling events
  wmmsens::processXevents();

  // adios!
  sensors_cleanup();  
  return 0;
}
